/************************************************************************/
/*   DIRUSE.C - directory usage display                                 */
/* -------------------------------------------------------------------- */
/*   This utility will display the disk usage from a given starting     */
/*   point and shows the biggest files. It sums up the files lengths of */
/*   all the files under this point.                                    */
/*                                                                      */
/*   To operate: DIRUSE [starting_path]                                 */
/*                                                                      */
/*   Written by:                                                        */
/*   Yaki Broshy, ISRAEL. Compuserve ID: 100274,1640                    */
/*                                                                      */
/*   Theory of operation:                                               */
/*   The main function search from starting level for all files.        */
/*   Starting level is the current dir if not overiden by command line. */
/*                                                                      */
/*   When directory is found, the pointer is advanced to new record and */
/*   function scandir is called to search and sum its sub directories.  */
/*   Function scandir gets a string with path name and recursivly       */
/*   searches downwards in the directories tree and sum all its files.  */
/*   Function show_usage display pie chart, and results list.           */
/*                                                                      */
/*   DIRUSE is written for Borland C++/Turbo C++ compiler.              */
/************************************************************************/

#include <graphics.h>
#include <stdlib.h>
#include <dos.h>
#include <conio.h>
#include <stdio.h>
#include <dir.h>
#include <math.h>
#include <string.h>
#include <ctype.h>

#define ARRAY_SIZE 700                  /* nax. number of directores */

void show_usage(void);                  /* prints results to screen */
void scandir(char *work_dir);           /* recursive searching downward */
int revcmp(const void *a, const void *b);

struct dir_ent  {
    long size;
    char name[MAXPATH];
} dir_array[ARRAY_SIZE];

struct dir_ent file_array[10];              /* top 10 big files */

struct dir_ent *pdir_array = &dir_array[0]; /* pointer to advance dir_array */
struct dir_ent *start = &dir_array[0];      /* pointer to keep start position */
char   root_dir[MAXPATH];                   /* save starting root dir */
char   start_dir[MAXPATH];                  /* start searching path */

/************************  add_file_to_list  ******************************/
void add_file_to_list(struct ffblk ffblk, char *dir_in_use)
{
    int i, j;

    if (ffblk.ff_fsize > file_array[9].size) {
        for (i=0; i<10; i++) {              // simple sort
            if (ffblk.ff_fsize > file_array[i].size) {
                for (j=9; j>i; j--) {       // shift lower down
                    file_array[j].size = file_array[j-1].size;
                    strcpy(file_array[j].name, file_array[j-1].name);
                }
                file_array[i].size = ffblk.ff_fsize;
                strcpy(file_array[i].name, dir_in_use);
                strcat(file_array[i].name, ffblk.ff_name);
                break;
            }
        }
    }
}

/*****************************  scandir  **********************************/
void scandir(char *work_dir)
{
    struct ffblk ffblk;
    int done, i, j;
    char temp[MAXPATH], disp[80];

    if (work_dir[ strlen( work_dir) -1] != '\\')
        strcat( work_dir, "\\");
    memset(disp, ' ', 80);
    disp[79] = '\0';
    strcpy(temp, work_dir);

    strncpy(disp, work_dir, strlen(work_dir));      // show current dir
    printf("\r%s", disp);
    strcat(work_dir, "*.*");
    done = findfirst( work_dir ,&ffblk, 0x3F);

    if (done==-1)
        return;

    while (!done) {
        if (! (ffblk.ff_name[0] == '.')) {
            if (ffblk.ff_attrib & FA_DIREC) {
                strcpy( work_dir,temp);
                strcat( work_dir, ffblk.ff_name);
                scandir( work_dir);
                strcpy( work_dir, temp);
            }
            else {
                pdir_array->size += ffblk.ff_fsize;
                add_file_to_list(ffblk, temp);
            }
        }
        done = findnext(&ffblk);
    }
}

/*****************************  revcmp  *******************************/
/* compare routine for sorting results */
int revcmp( const void *a, const void *b)
{
    struct dir_ent *x=(struct dir_ent *)a, *y=(struct dir_ent *)b;
    if (x->size < y->size)
        return(1);
    else if (x->size > y->size)
        return(-1);
    else
        return(0);

}


/******************************  show_usage  ****************************/
void show_usage()
{
    long bigsum=0, sum=0, disk_free=0, disk_size=0, disk_id=0;
    struct diskfree_t free;
    int gdriver = DETECT, gmode, errorcode;     /* request auto detection */
    int midx, midy, texth, starttxt;            /* mid of pie-chart */
    float startangle=0, angle, endangle, midangle, radius; /* for pie-slice */
    int  maxdir=0, i, color=1, pos;
    char tmp[80], *p;
    struct dfree dfree;


    if (registerbgidriver(EGAVGA_driver) < 0) {
        if (registerbgidriver(Herc_driver) < 0) {
            printf ("\nDIRUSE needs an EGA, VGA or hercules adapters.\n");
            exit(1);
        }
    }

    initgraph(&gdriver, &gmode, "");    /* initialize graphic mode */
    errorcode = graphresult();          /* read result of initialization */

    if (errorcode != grOk) {            /* an error occurred */
        printf("Graphics error: %s\n", grapherrormsg(errorcode));
        printf("Press any key to halt:");
        getch();
        exit(1);             /* return with error code */
    }
    starttxt = getmaxy() * 2 / 3;
    lineto(getmaxx(), 0);
    lineto(getmaxx(), getmaxy());
    lineto(0, getmaxy());
    lineto(0, 0);
    line(0, starttxt, 15, starttxt);
    line(25+textwidth("Biggest files"), starttxt, getmaxx(), starttxt);
    outtextxy(20, starttxt-textheight("A")/2, "Biggest files");

/* calculate  the sum off all size fileds in array */
    maxdir = 0;
    while (dir_array[maxdir].name[0])           // get total size
        bigsum += dir_array[maxdir++].size;

/* sort array */
    qsort(&dir_array[0], maxdir, sizeof(struct dir_ent), revcmp);

    maxdir = 0;
    while (dir_array[maxdir].name[0]) {
        sum += dir_array[maxdir++].size;
        if (100.0 * dir_array[maxdir-1].size / bigsum < 2)  // skip <2% directories
            break;
    }

/* display general info */
    settextstyle(DEFAULT_FONT, HORIZ_DIR, 2);
    texth = 1.5 * textheight("A");

    sprintf(tmp, "dir:  %s", root_dir);
    outtextxy(20, texth, tmp);
    texth += 1.5 * textheight("A");
    settextstyle(DEFAULT_FONT, HORIZ_DIR, 1);

    pos = 1 + strlen(root_dir);             // remember current working len
    if (root_dir[strlen(root_dir)-1] == '\\')
        pos --;
    sprintf(tmp, "files: %5.1f MB", 1.0 * bigsum / 1048576.0);
    outtextxy(20, texth, tmp);
    texth += 1.5 * textheight("A");


    if ((p = strchr(start_dir, ':')) != NULL)
        disk_id = toupper(*(p-1)) - 'A' + 1;
    _dos_getdiskfree(disk_id, &free);
    disk_free = (long) free.avail_clusters
                * (long) free.bytes_per_sector
                * (long) free.sectors_per_cluster;

    disk_size = (long) free.total_clusters
                * (long) free.bytes_per_sector
                * (long) free.sectors_per_cluster;

//  getdfree(getdisk()+1,&dfree);
//  disk_free = (long)dfree.df_avail * (long)dfree.df_bsec * (long)dfree.df_sclus;
//  disk_size = (long)dfree.df_total * (long)dfree.df_bsec * (long)dfree.df_sclus;

    sprintf(tmp, "free:  %5.1f MB  %3.0f%",
                  1.0 * disk_free / 1048576.0,
                  100.0 * disk_free / disk_size);
    outtextxy(20, texth, tmp);
    texth += 1.5 * textheight("A");

    sprintf(tmp, "total: %5.1f MB", 1.0 * disk_size / 1048576.0);
    outtextxy(20, texth, tmp);
    texth += 1.5 * textheight("A");

    texth += 1.5 * textheight("A");

/* set fill style and draw a pie slice */
    midy    = getmaxy() /3;
    radius  = midy * 5 / 7;
    midx    = getmaxx() - radius - 8 * textwidth("1");

    for (i=0; i<maxdir; i++) {
        angle = (dir_array[i].size * 360.0) / sum;
        endangle = startangle + angle;
        midangle = (startangle + angle/2) / 180 * 3.141592654;
        setfillstyle(SOLID_FILL, color);
        setcolor(WHITE);
        pieslice( midx, midy, startangle, endangle, radius);
        sprintf(tmp, "%3.0f%%", 100.0 * dir_array[i].size / sum);
        moveto(midx + radius * cos(midangle)*1.15,
               midy - radius * sin(midangle)*1.15);
        moverel(-1 * textwidth(tmp)/2 - 5,  -1 * textheight(tmp)/2);
        outtext(tmp);

        if (dir_array[i].size > 104857L)
            sprintf(tmp, "%6.1f MB  %3.0f%%   %s",
                    dir_array[i].size / 1048576.0,
                    100.0 * dir_array[i].size / sum,
                    (pos == strlen(dir_array[i].name)) ?
                    "current dir" : &dir_array[i].name[pos]);
        else
            sprintf(tmp, "%6.1f KB  %3.0f%%   %s",
                    dir_array[i].size / 1024.0,
                    100.0 * dir_array[i].size / sum,
                    (pos == strlen(dir_array[i].name)) ?
                    "current dir" : &dir_array[i].name[pos]);

        if (texth + (i+2) * 2 * textheight("A") < starttxt) {
            setcolor(color);
            outtextxy( 20,  texth + (i+1) * 2 * textheight("A"), "");
            setcolor(WHITE);
            outtextxy( 20 + textwidth(""), texth + (i+1) * 2 * textheight("A"), tmp);
        }
        color++;
        if (color == getmaxcolor())
           setcolor(1);
        if (color > getmaxcolor())
            color = 1;

        startangle = endangle;
    }

    setcolor(WHITE);
    for (i=0; i<10; i++) {              // show biggest files
        if (file_array[i].size > 104857L)
            sprintf(tmp, "%2d %6.1f MB  %3.0f%%   %s",
                    i + 1,
                    file_array[i].size / 1048576.0,
                    100.0 * file_array[i].size / sum,
                    file_array[i].name);
        else
            if (file_array[i].size > 0L)
                sprintf(tmp, "%2d %6.1f KB  %3.0f%%   %s",
                    i + 1,
                    file_array[i].size / 1024.0,
                    100.0 * file_array[i].size / sum,
                    file_array[i].name);
            else
                break;
        outtextxy( 20 + textwidth(""), starttxt + (i+1) * 2 * textheight("A"), tmp);
    }

/* clean up */
   getch();
   closegraph();
}

/******************************  main  *********************************/
void main(int argc, char *argv[])
{
    struct ffblk ffblk;
    int i, done;
    char temp[MAXPATH], scratch[MAXPATH];

    if (argc == 2)
        strcpy(start_dir, argv[1]);
    else
        getcwd(start_dir, MAXPATH);

    for (i=0; i<strlen(start_dir); i++)
        start_dir[i] = toupper(start_dir[i]);
    if (start_dir[ strlen( start_dir) -1] != '\\')  /* nead this for root dir */
        strcat( start_dir, "\\");
    strcpy(root_dir, start_dir);
    strcpy(temp, start_dir);        /* to keep starting path level */
    strcpy( pdir_array->name, start_dir);
    pdir_array++;
    printf("\n%s", start_dir);
    strcpy(scratch, start_dir);
    strcat(scratch, "*.*");

    done = findfirst(scratch, &ffblk, 0x3F);
    if (done == -1) {
        printf("\nInvalid path\n");
        exit(1);
    }
    while ( !done) {                    /* scan first level dir's */
        if (ffblk.ff_name[0] != '.') {              /* skip two hidden files */
            if (ffblk.ff_attrib & FA_DIREC) {       /* if its a directory */
                strcat(start_dir, ffblk.ff_name);   /* add its path */
                strcpy(pdir_array->name, start_dir);/* its name to record */
                scandir(start_dir);                 /* recursive call here */
                strcpy(start_dir, temp);            /* restore start path */
                pdir_array++;
            }
            else {
                start->size += ffblk.ff_fsize;      /* sum the files */
                add_file_to_list(ffblk, temp);
            }
        }
        done = findnext(&ffblk);
    }
    show_usage();                                   /* pie & results */
}

